use lazy_static::lazy_static;
use alloc::vec::Vec;
use crate::config::MAX_APP_NUM;
use crate::sync::UPSafeCell;
use crate::task::task::{TaskControlBlock, TaskStatus};
use crate::loader::{get_num_app, get_app_data};
use crate::println;
use crate::task::context::TaskContext;
use crate::task::switch::__switch;
use crate::timer::{get_time_ms, get_time_us};
use crate::trap::TrapContext;

mod context;
mod switch;
mod task;


// 公开到外部的一个全局任务管理器的结构体
pub struct TaskManager {
    num_app: usize,  // app的数量 这个在os运行之后就不会有变化
    inner: UPSafeCell<TaskManagerInner>,  // 这个内部TaskControlBlock 会随着系统运行发生变化
}

impl TaskManager {
    // 即将进入用户态, 把之前使用的内核时间统计
    pub fn user_time_start(&self){
        let mut inner = self.inner.exclusive_access();
        let current_task_id = inner.current_task_id;
        let kernel_use_time = inner.refresh_stop_clock();

        inner.tasks[current_task_id].kernel_time += kernel_use_time;
    }

    pub fn kernel_time_start(&self){
        let mut inner = self.inner.exclusive_access();
        let current_task_id = inner.current_task_id;
        let user_use_time = inner.refresh_stop_clock();

        inner.tasks[current_task_id].user_time += user_use_time;
    }


    fn run_first_task(&self){
        // 得到下一个需要执行的任务,的上下文 这里我们由于第一次执行, 下一个任务我们指定为0下标的任务即可
        let next_task_context_ptr = {
            let mut inner = self.inner.exclusive_access();
            // 第一次掐表, 开始记录时间
            inner.refresh_stop_clock();

            let task_0 = &mut inner.tasks[0];
            // 修改任务0的状态
            task_0.task_status = TaskStatus::Running;

            &(task_0.task_cx) as *const TaskContext
        };

        // 由于第一次切换, 我们current task context 不存在, 所以我们手动创建一个, 在调用switch的时候
        // 这里调用switch的状态会被保存在unused 这里, 但是一旦程序开始运行, 就开始在内核栈或者用户栈之间切换, 永远
        // 不会走到这里了
        let mut unused = TaskContext::new();

        println!("start run app");
        unsafe {
            __switch(&mut unused as *mut TaskContext, next_task_context_ptr)
        }

        //
        panic!("unreachable in run_first_task!");
    }

    // 停止当前正在运行的任务
    fn mark_current_exit(&self){
        let mut inner = self.inner.exclusive_access();
        let current_task_id = inner.current_task_id;
        let kernel_use_time = inner.refresh_stop_clock();
        let current_task = &mut inner.tasks[current_task_id];
        current_task.task_status = TaskStatus::Exited;

        // 统计内核时间, 并输出这个任务 花费了多少内核时间
        current_task.kernel_time += kernel_use_time;
        println!("task {:?}, exit, kernel_time: {:?}ms, user_time: {:?}ms", current_task_id, current_task.kernel_time/1000, current_task.user_time/1000);
    }

    // 挂起当前任务
    fn mark_current_suspend(&self){
        let mut inner = self.inner.exclusive_access();
        let current_task_id = inner.current_task_id;
        let kernel_use_time = inner.refresh_stop_clock();
        let current_task = &mut inner.tasks[current_task_id];
        current_task.task_status = TaskStatus::Ready;

        // 统计内核时间(因为内核进入到这里之前肯定是掐表了)
        // 挂起任务A之后,这里也掐表了, 切到新的任务B, B开始执行, B在挂起的时候 也会走入到这里掐表, 顺便并得到任务A掐表到现在任务B掐表准备切换任务出去中间的时间
        current_task.kernel_time += kernel_use_time;
    }

    // 运行下一个任务
    fn run_next_task(&self){
        if let Some(next_task_id) = self.find_next_task() {
            // 得到当前任务context ptr 以及 下一个任务context的ptr
            let (current_task_ptr,next_task_ptr) = {
                let mut inner = self.inner.exclusive_access();

                // 当前任务id
                let current_task_id = inner.current_task_id;

                // 修改current_task_id 为 下一个任务, 因为一旦到switch就是在运行下一个任务了
                inner.current_task_id = next_task_id;

                // 修改下一个任务的状态
                inner.tasks[next_task_id].task_status = TaskStatus::Running;

                // 得到current的ptr
                let current_task_ptr = &mut inner.tasks[current_task_id].task_cx as *mut TaskContext;

                // 得到需要被切换的下一个任务ptr
                let next_task_ptr = &inner.tasks[next_task_id].task_cx as *const TaskContext;
                (current_task_ptr, next_task_ptr)
            };

            // 开始伟大的切换!
            unsafe {
                __switch(current_task_ptr, next_task_ptr);
            }
        } else {
            panic!("All applications completed!");
        }
    }

    // 找到一个除当前任务之外的 是Ready状态的任务的id
    fn find_next_task(&self) -> Option<usize>{
        let mut inner = self.inner.exclusive_access();
        let current = inner.current_task_id;
        for task_id in current + 1..current + self.num_app + 1 {
            let _tmp_id = task_id % self.num_app;
            if inner.tasks[_tmp_id].task_status == TaskStatus::Ready {
                return Some(_tmp_id)
            }
        }
        None
    }

    // 得到当前正在执行用户应用的 trap context
    pub fn get_current_trap_cx(&self) -> &'static mut TrapContext {
        let inner = self.inner.exclusive_access();
        inner.tasks[inner.current_task_id].get_trap_cx()
    }

    // 得到当前正在只用的用户应用的 页表
    pub fn get_current_token(&self) -> usize {
        let inner = self.inner.exclusive_access();
        inner.tasks[inner.current_task_id].get_user_token()
    }

    // 扩张/缩减, 当前应用的地址空间中的堆段, 指定字节的数据
    pub fn change_current_program_brk(&self, size: i32) -> Option<usize> {
        let mut inner = self.inner.exclusive_access();
        let current_task_id = inner.current_task_id;
        inner.tasks[current_task_id].change_program_brk(size)
    }
}


// 不公开的结构体, 有一个MAX_APP_NUM大小的数组, 用来保存TCB, 和当前任务的的TCB的下标
struct TaskManagerInner {
    tasks: Vec<TaskControlBlock>,
    current_task_id: usize,
    stop_clock_time: usize,  // 记录最近一次停表时间
}

impl TaskManagerInner {
    fn refresh_stop_clock(&mut self) -> usize {
        // 上次停表的时间
        let start_time = self.stop_clock_time;
        // 当前时间
        self.stop_clock_time = get_time_us();
        // 返回 当前时间 - 上次停表时间 = 时间间距
        self.stop_clock_time - start_time
    }
}

lazy_static! {
    pub static ref TASK_MANAGER: TaskManager = {
        println!("init TASK_MANAGER");
        let num_app = get_num_app();
        println!("num_app = {}", num_app);
        let mut tasks: Vec<TaskControlBlock> = Vec::new();

        for i in 0..num_app {
            // 得到elf 的二进制数据
            let app_elf_data = get_app_data(i);
            // 根据二进制数据创建 TCB
            let tcb = TaskControlBlock::from(app_elf_data, i);
            tasks.push(tcb);
        }
        TaskManager {
            num_app,
            inner: unsafe {
                UPSafeCell::new(TaskManagerInner {
                    tasks,
                    stop_clock_time: 0,
                    current_task_id: 0,  // 从第0个启动
                })
            },
        }
    };
}


// 由内核调用, 在main中, 用来 作为执行用户应用的开端
pub fn run_first_task() {
    TASK_MANAGER.run_first_task();
}

// 挂起当前任务, 并运行下一个
pub fn suspend_current_and_run_next(){
    TASK_MANAGER.mark_current_suspend();
    TASK_MANAGER.run_next_task();
}

// 退出当前任务, 并运行下一个
pub fn exit_current_and_run_next(){
    TASK_MANAGER.mark_current_exit();
    TASK_MANAGER.run_next_task();
}

// 扩张/缩减, 当前应用的地址空间中的堆段, 指定字节的数据
pub fn change_program_brk(size: i32) -> Option<usize> {
    TASK_MANAGER.change_current_program_brk(size)
}

// 得到当前运行任务的 trap context 的地址
pub fn current_trap_cx() -> &'static mut TrapContext {
    TASK_MANAGER.get_current_trap_cx()
}
